﻿using OpenBveApi.Runtime;

namespace OpenbveFcmbTrainPlugin
{
    /// <summary>A device simulating ATC by Bombardier.</summary>
    internal class AtcBombardier : Device
    {
        /// <summary>Represents the state of the device.</summary>
        private enum DeviceStates
        {
            /// <summary>The device is being overriden.</summary>
            Override,
            /// <summary>The device is initializing.</summary>
            Initializing,
            /// <summary>The device has been initialized and no driving mode is selected.</summary>
            Initialized,
            /// <summary>The device is in YARD (M+25) driving mode.</summary>
            YARD,
            /// <summary>The device is in ATP (M+ATP) driving mode.</summary>
            ATP,
            /// <summary>The device is in ATO mode.</summary>
            ATO,
        }

        /// <summary>The current state of the device.</summary>
        private DeviceStates DeviceState;

        /// <summary>Represents the state of the ATC control.</summary>
        private enum AtcControlStates
        {
            /// <summary>The brakes are released.</summary>
            Released,
            /// <summary>The power is cut.</summary>
            NoPower,
            /// <summary>The service brake is applied and can be released.</summary>
            BrakeService,
            /// <summary>The service brake is applied and cannot be released.</summary>
            BrakeServiceNoRelease,
            /// <summary>The service brake is fully applied.</summary>
            BrakeFull,
            /// <summary>The emergency brake is applied.</summary>
            BrakeEmergency,
        }

        /// <summary>The current state of the ATC control.</summary>
        private AtcControlStates AtcControlState;

        /// <summary>Represents the state of the track.</summary>
        private enum TrackStates
        {
            Unprotected,
            Enable,
            ATC,
            Disable
        }

        /// <summary>The current state of the track.</summary>
        private TrackStates TrackState;

        /// <summary>The time needed by the ATC device to initialize, in seconds.</summary>
        private double InitializationTime;

        /// <summary>A counter to keep track of initialization time, in seconds.</summary>
        private double InitializationCounter;

        /// <summary>The blinking time for lamps, in milliseconds.</summary>
        private double BlinkTime;

        /// <summary>A counter to keep track of blink time, in milliseconds.</summary>
        private double BlinkCounter;

        /// <summary>The time after which YARD mode enters idle state and engages the service brake, in seconds.</summary>
        private double YardIdleTime;

        /// <summary>A counter to keep track of YARD mode idle time, in seconds.</summary>
        private double YardIdleCounter;

        /// <summary>The maximum speed in YARD mode, in km/h.</summary>
        private double YardMaximumSpeed;

        /// <summary>Represents an ATC movement permission order.</summary>
        private class MovementPermission
        {
            internal Speed LimitBase { get; private set; }
            internal Speed LimitNoTraction { get; private set; }
            internal Speed LimitService { get; private set; }
            internal Speed LimitEmergency { get; private set; }
            internal Speed TargetLimit { get; private set; }
            internal double TargetLocation { get; private set; }

            internal MovementPermission(Speed limit, Speed target, double targetLocation)
            {
                LimitBase = limit;
                LimitNoTraction = new Speed(limit.MetersPerSecond + 3 / 3.6);
                LimitService = new Speed(limit.MetersPerSecond + 6 / 3.6);
                LimitEmergency = new Speed(limit.MetersPerSecond + 9 / 3.6);
                TargetLimit = target;
                TargetLocation = targetLocation;
            }
        }

        /// <summary>The current permission regarding speed and travel distance.</summary>
        private MovementPermission CurrentPermission;

        /// <summary>The ideal power notch for the AI.</summary>
        private int AiPowerNotch;

        /// <summary>The ideal brake notch for the AI.</summary>
        private int AiBrakeNotch;

        /// <summary>Creates an instance of the Bombardier ATC device.</summary>
        /// <param name="initTime">The time needed by the ATC device to initialize, in seconds.</param>
        /// <param name="yardIdleTime">The time after which YARD mode enters idle state and engages the service brake, in seconds.</param>
        /// <param name="yardMaxSpeed">The maximum speed in YARD mode, in km/h.</param>
        /// <param name="blinkTime">The blinking time for lamps, in milliseconds.</param>
        internal AtcBombardier(double initTime, double yardIdleTime, double yardMaxSpeed, double blinkTime)
        {
            InitializationTime = initTime;
            YardIdleTime = yardIdleTime;
            YardMaximumSpeed = yardMaxSpeed;
            BlinkTime = blinkTime;
        }

        /// <summary>Is called when the device state should be updated.</summary>
        /// <param name="train">The current train.</param>
        /// <param name="route">The current route.</param>
        /// <param name="init">Whether the device should initialize.</param>
        /// <param name="elapsedTime">The time elapsed since the previous call.</param>
        internal override void Update(Train train, Route route, bool init, Time elapsedTime)
        {
            if (init)
            {
                // Initialize device on start
                DeviceState = DeviceStates.Initializing;
            }

            // Update ATC movement permission
            UpdatePermission();

            double speed = train.State.Speed.KilometersPerHour;
            bool stopped = speed < 0.05;

            // Speed limit enforcement
            if (DeviceState == DeviceStates.YARD || DeviceState == DeviceStates.ATP || DeviceState == DeviceStates.ATO)
            {
                if (AtcControlState == AtcControlStates.Released && speed > CurrentPermission.LimitNoTraction.KilometersPerHour)
                {
                    // Cut power on power limit
                    AtcControlState = AtcControlStates.NoPower;
                }
                if (AtcControlState == AtcControlStates.NoPower && speed < CurrentPermission.LimitNoTraction.KilometersPerHour)
                {
                    // Stop cutting power below power limit
                    AtcControlState = AtcControlStates.Released;
                }
                if (speed > CurrentPermission.LimitService.KilometersPerHour && speed < CurrentPermission.LimitEmergency.KilometersPerHour)
                {
                    // Apply service brake on limit
                    AtcControlState = AtcControlStates.BrakeServiceNoRelease;
                }
                if (speed > CurrentPermission.LimitEmergency.KilometersPerHour)
                {
                    // Unselect driving mode to apply emergency brake if overspeeding
                    DeviceState = DeviceStates.Initialized;
                }

                // Allow brake release under defined threshold
                if (AtcControlState == AtcControlStates.BrakeServiceNoRelease && speed < CurrentPermission.LimitBase.KilometersPerHour)
                {
                    AtcControlState = AtcControlStates.BrakeService;
                }
            }

            // Brake application
            switch (AtcControlState)
            {
                case AtcControlStates.Released:
                    RequestedBrakeNotch = -1;
                    RequestedPowerNotch = -1;
                    break;
                case AtcControlStates.NoPower:
                    RequestedBrakeNotch = -1;
                    RequestedPowerNotch = 0;
                    break;
                case AtcControlStates.BrakeService:
                case AtcControlStates.BrakeServiceNoRelease:
                    RequestedBrakeNotch = train.ServiceBrakeNotch;
                    RequestedPowerNotch = 0;
                    break;
                case AtcControlStates.BrakeEmergency:
                    RequestedBrakeNotch = train.Specs.BrakeNotches + 1;
                    RequestedPowerNotch = 0;
                    break;
            }

            switch (DeviceState)
            {
                // ATC device is disabled (also called "Special Mode")
                case DeviceStates.Override:
                    train.ContinuousProtection = false;
                    train.VigilanceOverride = false;
                    // Release control of the brakes and power
                    AtcControlState = AtcControlStates.Released;
                    break;
                // ATC device is initializing
                case DeviceStates.Initializing:
                    train.ContinuousProtection = false;
                    train.VigilanceOverride = false;
                    // Apply emergency brake during initialization
                    AtcControlState = AtcControlStates.BrakeEmergency;
                    // Wait during initialization time
                    InitializationCounter += elapsedTime.Seconds;
                    // If initialization is complete, change the state of the device
                    if (InitializationCounter > InitializationTime)
                    {
                        DeviceState = DeviceStates.Initialized;
                        InitializationCounter = 0;
                    }
                    break;
                // ATC device is initialized and a driving mode is required
                case DeviceStates.Initialized:
                    train.ContinuousProtection = false;
                    train.VigilanceOverride = false;
                    // Apply service/emergency brake while no driving mode is selected
                    AtcControlState = stopped ? AtcControlStates.BrakeServiceNoRelease : AtcControlStates.BrakeEmergency;
                    // ATC movement permission
                    CurrentPermission = new MovementPermission(new Speed(0), new Speed(0), 0);
                    break;
                // ATC device is in YARD (M+25) driving mode
                case DeviceStates.YARD:
                    train.ContinuousProtection = false;
                    train.VigilanceOverride = false;
                    // If the train is not moving, count idle time
                    if (stopped)
                    {
                        YardIdleCounter += elapsedTime.Seconds;
                        // If above idle time, apply the service brake
                        AtcControlState = YardIdleCounter >= YardIdleTime ? AtcControlStates.BrakeService : AtcControlStates.Released;
                    }
                    else
                    {
                        YardIdleCounter = 0;
                    }
                    // Apply brake if any door opens
                    if (train.DoorState != DoorStates.None)
                    {
                        RequestedBrakeNotch = stopped ? train.Specs.BrakeNotches + 1 : train.ServiceBrakeNotch;
                    }
                    break;
                // ATC device is in ATP (M+ATP) driving mode
                case DeviceStates.ATP:
                    train.ContinuousProtection = true;
                    train.VigilanceOverride = false;
                    break;
                // ATC device is in ATO driving mode
                case DeviceStates.ATO:
                    train.ContinuousProtection = true;
                    train.VigilanceOverride = true;
                    break;
            }

            // Blinking light animation
            BlinkCounter += elapsedTime.Milliseconds;
            bool blink = BlinkCounter >= BlinkTime;
            if (BlinkCounter > (BlinkTime * 2)) BlinkCounter = 0;

            // Panel indicators
            train.Panel[25] = DeviceState == DeviceStates.Override ? 1 : 0;
            train.Panel[26] = DeviceState == DeviceStates.YARD ? 1 : 0;
            train.Panel[26] += DeviceState == DeviceStates.Initialized && stopped && blink ? 1 : 0;
            train.Panel[27] = DeviceState == DeviceStates.ATP ? 1 : 0;
            train.Panel[28] = DeviceState == DeviceStates.ATP ? 1 : 0;
            switch (AtcControlState)
            {
                case AtcControlStates.Released:
                case AtcControlStates.NoPower:
                    train.Panel[32] = 0;
                    break;
                case AtcControlStates.BrakeServiceNoRelease:
                case AtcControlStates.BrakeFull:
                case AtcControlStates.BrakeEmergency:
                    train.Panel[32] = DeviceState != DeviceStates.Initializing ? 1 : 0;
                    break;
                case AtcControlStates.BrakeService:
                    train.Panel[32] = blink ? 1 : 0;
                    break;
            }
            train.Panel[34] = (int)CurrentPermission.LimitBase.KilometersPerHour * 1000;
        }


        /// <summary>Is called when the state of a key changes.</summary>
        /// <param name="key">The key.</param>
        /// <param name="pressed">Whether the key is pressed or released.</param>
        /// <param name="train">The current train.</param>
        internal override void KeyChange(VirtualKeys key, bool pressed, Train train)
        {
            double speed = train.State.Speed.KilometersPerHour;
            bool stopped = speed < 0.05;

            if (pressed)
            {
                switch (key)
                {
                    case VirtualKeys.I:
                        // YARD (M+25) mode selection button
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                            if (DeviceState == DeviceStates.Initialized && stopped)
                            {
                                DeviceState = DeviceStates.YARD;
                            }
                        }
                        break;
                    case VirtualKeys.J:
                        // ATP (M+ATP) mode selection button
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                        }
                        break;
                    case VirtualKeys.K:
                        // ATO mode selection button
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                        }
                        break;
                    case VirtualKeys.L:
                        // ATO start button
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                        }
                        break;
                    case VirtualKeys.B1:
                        // ATC brake release
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                            // Allow brake release under certain conditions
                            if (AtcControlState == AtcControlStates.BrakeService)
                            {
                                AtcControlState = AtcControlStates.Released;
                                // Reset YARD idle counter
                                YardIdleCounter = 0;
                            }
                        }
                        break;
                    case VirtualKeys.C1:
                        // Override toggle
                        if (!OpenbveFcmbTrainPlugin.KeysPressed[(int)key])
                        {
                            switch (DeviceState)
                            {
                                case DeviceStates.Initializing:
                                case DeviceStates.Initialized:
                                case DeviceStates.YARD:
                                    // Override device if possible
                                    if (stopped && train.PhysicalHandles.PowerNotch == 0)
                                    {
                                        DeviceState = DeviceStates.Override;
                                    }
                                    break; 
                                case DeviceStates.Override:
                                    // Disable override if possible
                                    if (stopped && train.PhysicalHandles.PowerNotch == 0)
                                    {
                                        DeviceState = DeviceStates.Initializing;
                                    }
                                    break;
                            }
                        }
                        break;
                }
            }
        }

        /// <summary>Is called to inform about signals.</summary>
        /// <param name="signal">The signal data.</param>
        internal override void SetSignal(SignalData[] signal)
        {

        }

        /// <summary>Is called when a beacon is passed.</summary>
        /// <param name="beacon">The beacon data.</param>
        internal override void SetBeacon(BeaconData beacon)
        {
            int type = beacon.Type;
            switch (type)
            {
                case -16777215:
                    {
                        if (!(beacon.Optional >= 0 & beacon.Optional <= 3))
                        {
                            break;
                        }
                        TrackState = (TrackStates)beacon.Optional;
                        return;
                    }
            }
        }

        /// <summary>Is called when the device should perform the AI.</summary>
        /// <param name="data">The AI data.</param>
        /// <param name="train">The current train.</param>
        /// <param name="route">The current route.</param>
        internal override void PerformAI(AIData data, Train train, Route route)
        {
            // Set AI notches to what the AI is doing right now
            AiBrakeNotch = data.Handles.BrakeNotch;
            AiPowerNotch = data.Handles.PowerNotch;

            double speed = train.State.Speed.KilometersPerHour;
            bool stopped = speed < 0.05;

            switch (DeviceState)
            {
                case DeviceStates.Initializing:
                    data.Handles.BrakeNotch = train.Specs.BrakeNotches;
                    data.Response = AIResponse.Short;
                    break;
                case DeviceStates.Initialized:
                    // Apply brake
                    data.Handles.BrakeNotch = train.Specs.BrakeNotches;
                    data.Response = AIResponse.Short;
                    // If the train is stopped, select YARD (M+25) mode
                    if (stopped)
                    {
                        KeyChange(VirtualKeys.I, true, train);
                        data.Response = AIResponse.Short;
                        KeyChange(VirtualKeys.I, false, train);
                    }
                    break;
                case DeviceStates.YARD:
                case DeviceStates.ATP:
                    // If the ATC brakes can be released, press the ATC brake release button
                    if (AtcControlState == AtcControlStates.BrakeService)
                    {
                        KeyChange(VirtualKeys.B1, true, train);
                        data.Response = AIResponse.Medium;
                        KeyChange(VirtualKeys.B1, false, train);
                    }
                    // Calculate ideal notch for AI when approaching the speed limit
                    if (speed > CurrentPermission.LimitBase.KilometersPerHour - 2 && train.Acceleration > 0.1)
                    {
                        if (data.Handles.PowerNotch > 0)
                        {
                            AiPowerNotch -= 1;
                        }
                        else if (train.Specs.HasHoldBrake)
                        {
                            data.Handles.HoldBrake = true;
                        }
                        else
                        {
                            AiBrakeNotch += 1;
                        }
                    }
                    else if (speed < CurrentPermission.LimitBase.KilometersPerHour - 4 && train.Acceleration < 0.25)
                    {
                        if (data.Handles.BrakeNotch > 0)
                        {
                            AiBrakeNotch -= 1;
                        }
                        else
                        {
                            AiPowerNotch += 1;
                        }
                    }
                    // If still far from next station, take control of AI handles
                    if (speed / (route.CurrentStation.StopPosition - train.State.Location) < 0.2)
                    {
                        if (speed > CurrentPermission.LimitBase.KilometersPerHour - 5)
                        {
                            data.Handles.PowerNotch = AiPowerNotch;
                            data.Handles.BrakeNotch = AiBrakeNotch;
                            data.Response = AIResponse.Medium;
                        }
                    }
                    break;
            }
        }

        /// <summary>Updates the data for the current and next station.</summary>
        internal void UpdatePermission()
        {
            Speed zero = new Speed(0);
            Speed yard = new Speed(YardMaximumSpeed / 3.6);

            switch (DeviceState)
            {
                case DeviceStates.Override:
                case DeviceStates.Initializing:
                case DeviceStates.Initialized:
                    if (CurrentPermission == null || CurrentPermission.LimitBase != zero || CurrentPermission.TargetLimit != zero)
                    {
                        CurrentPermission = new MovementPermission(zero, zero, 0);
                    }
                    break;
                case DeviceStates.YARD:
                    if (CurrentPermission == null || CurrentPermission.LimitBase != yard || CurrentPermission.TargetLimit != yard)
                    {
                        CurrentPermission = new MovementPermission(yard, yard, 0);
                    }
                    break;
                case DeviceStates.ATP:
                case DeviceStates.ATO:
                    if (CurrentPermission == null || CurrentPermission.LimitBase != zero || CurrentPermission.TargetLimit != zero)
                    {
                        CurrentPermission = new MovementPermission(zero, zero, 0);
                    }
                    break;
            }
        }

    }
}
